import re
from typing import (
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from configmate.base import constants, operators, types

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)


###
# Filters for command line arguments
###
class CliSectionReader(operators.Operator[Collection[str], List[str]]):
    """Reads a section of command line arguments."""

    def __init__(
        self,
        section_start_token: Optional[str] = None,
        section_end_token: Optional[str] = constants.CLI_SECTION_END_TOKEN,
    ) -> None:
        super().__init__()
        self.start_token = section_start_token
        self.end_token = section_end_token

    def _transform(self, ctx: operators.Context, input_: Collection[str]) -> List[str]:
        section = SectionTracker(self.start_token, self.end_token)
        return [arg for arg in input_ if arg in section]


class ArgSelector(operators.Operator[Sequence[str], List[str]]):
    """Filters out command line arguments using a callable."""

    def __init__(self, arg_prefix: str = constants.CLI_OVERLAY_FILE_PREFIX) -> None:
        super().__init__()
        self._prefix = arg_prefix
        self._trigger = PrefixTrigger(self._prefix)

    def _transform(self, ctx: operators.Context, input_: Sequence[str]) -> List[str]:
        return [arg.lstrip(self._prefix) for arg in input_ if self._trigger(arg)]


class KeyValueSelector(operators.Operator[Sequence[str], List[Tuple[str, str]]]):
    """Filters out command line arguments using a callable."""

    def __init__(
        self, key_prefix: str = constants.CLI_OVERLAY_KWARG_KEY_PREFIX
    ) -> None:
        super().__init__()
        self._prefix = key_prefix
        self._trigger = PrefixTrigger(key_prefix)

    def _transform(
        self, ctx: operators.Context, input_: Sequence[str]
    ) -> List[Tuple[str, str]]:
        pairs = zip(input_, input_[1:])
        return [(k.lstrip(self._prefix), v) for k, v in pairs if self._trigger(k)]


class DictParser(operators.Operator[Tuple[str, str], types.NestedDict[T_co]]):
    """Filters out command line arguments using a callable."""

    def __init__(
        self,
        key_split_delimiter=constants.CLI_OVERLAY_KWARG_KEY_DELIMITER,
        arg_parser: Callable[[str], T_co] = constants.CLI_OVERLAY_KWARG_VAL_PARSER,
    ) -> None:
        super().__init__()
        self._delimiter = key_split_delimiter
        self._parser = arg_parser

    def _transform(
        self, ctx: operators.Context, input_: Tuple[str, str]
    ) -> types.NestedDict[T_co]:
        key, value = input_
        return nest_in_dict(key.split(self._delimiter), self._parser(value))


###
# Helpers
###
def lstrip_prefix(prefix: str, string: str) -> str:
    """Removes the prefix from the string."""
    return string[len(prefix) :]


def nest_in_dict(key_sequence: Iterable[str], value: T) -> types.NestedDict[T]:
    """Make a nested dictionary from a sequence of keys and a value.

    E.g. nest_in_dict(["a", "b", "c"], 1) -> {"a": {"b": {"c": 1}}}
    """

    def nest_dict() -> Union[Dict, T]:
        key = next(key_iterator, types.SENTINEL)
        return {key: nest_dict()} if key != types.SENTINEL else value

    key_iterator = iter(key_sequence)
    return cast(types.NestedDict[T], nest_dict())


class SectionTracker:
    """Tracks whether a section has started and ended.

    A section starts when the first argument matches the start trigger.
    A section ends when the first argument matches the end trigger.
    """

    def __init__(self, start: Optional[str], end: Optional[str]) -> None:
        self._start_trigger = PrefixTrigger(start) if start else lambda _: True
        self._end_trigger = PrefixTrigger(end) if end else lambda _: False
        self._section_started = self._section_ended = False

    def __contains__(self, arg: str) -> bool:
        self._section_started |= self._start_trigger(arg)  # type: ignore
        self._section_ended |= self._end_trigger(arg)  # type: ignore
        return self._section_started and not self._section_ended


class PrefixTrigger:
    """A callable that returns True if the argument starts with the prefix."""

    def __init__(self, prefix: str) -> None:
        escaped_prefix = re.escape(prefix)
        self.pattern = re.compile(f"^{escaped_prefix}([^{escaped_prefix}]|$)")

    def __call__(self, arg: str) -> bool:
        return self.pattern.search(arg) is not None



import abc
import collections
import copy
import itertools
import types
from typing import (
    Callable,
    DefaultDict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    overload,
)
from typing_extensions import Self

T = TypeVar("T")
U = TypeVar("U")
T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)


class Context:
    "Holds the context for the configmate pipeline."
    _namespaces: DefaultDict[int, types.SimpleNamespace]
    _children: List["Context"]

    def __init__(self) -> None:
        self._namespaces = collections.defaultdict(types.SimpleNamespace)
        self._children = []

    def __getitem__(self, key: object) -> types.SimpleNamespace:
        return self._namespaces[id(key)]

    def __copy__(self) -> "Context":
        new = Context()
        new._namespaces = copy.copy(self._namespaces)
        return new

    def get_child(self) -> "Context":
        child = copy.copy(self)
        self._children.append(child)
        return child


Callback = Callable[[Context, T_contra], None]


class Operator(abc.ABC, Generic[T_contra, T_co]):
    input_type: Type[T_contra] = object  # type: ignore
    output_type: Type[T_co] = object  # type: ignore

    @abc.abstractmethod
    def _transform(self, ctx: Context, input_: T_contra) -> T_co:
        """Apply the transformation to the input
        has access to the context via `self.context`
        """

    def __init__(self, *_, **__) -> None:
        super().__init__()
        self._callbacks: List[Callback[T_co]] = []

    def __call__(self, input_: T_contra, _ctx: Optional[Context] = None) -> T_co:
        _ctx = Context() if _ctx is None else _ctx
        result = self._transform(_ctx, input_)
        self._run_callbacks(_ctx, result)
        return result

    # fmt: off
    @overload
    def pipe_to(self, step: None) -> Self: ...
    @overload
    def pipe_to(self, step: "Operator[T_co, T]") -> "Pipeline[T_contra, T]": ...
    # fmt: on
    def pipe_to(self, step: "Optional[Operator]") -> "Operator":
        return self if step is None else Pipeline(self, step)

    def append_callback(self, callback: Callback[T_co]) -> "Operator[T_contra, T_co]":
        self._callbacks.append(callback)
        return self

    def _run_callbacks(self, ctx: Context, result) -> None:
        for callback in self._callbacks:
            callback(ctx, result)


class Pipeline(Operator[T_contra, T_co]):
    """A pipeline of operators that are applied in sequence."""

    def __init__(self, first: Operator[T_contra, T], second: Operator[T, T_co]) -> None:
        super().__init__()
        self._first = first
        self._next = second

    def _transform(self, ctx: Context, input_: T_contra) -> T_co:
        return self._next(self._first(input_, ctx), ctx)

    @property
    def input_type(self) -> Type[T_contra]:  # type: ignore
        return self._first.input_type

    @property
    def output_type(self) -> Type[T_co]:  # type: ignore
        return self._next.output_type


class MapIterable(Operator[Iterable[T_contra], Iterator[T_co]]):
    """Maps an ~Operator over an iterable."""

    input_type = Iterable  # type: ignore
    output_type = Iterator  # type: ignore

    def __init__(self, map_step: Operator[T_contra, T_co]) -> None:
        super().__init__()
        self._map_step = map_step

    def _transform(self, ctx: Context, input_: Iterable[T_contra]) -> Iterator[T_co]:
        return (self._map_step(i, ctx.get_child()) for i in input_)


class JoinOutputs(Operator[T_contra, Iterator[T_co]]):
    """Joins the outputs of multiple operators into a single iterable."""

    output_type = Iterator  # type: ignore

    def __init__(self, *steps: Operator[T_contra, T_co]) -> None:
        super().__init__()
        self._steps = steps

    def _transform(self, ctx: Context, input_: T_contra) -> Iterator[T_co]:
        return (step(input_, ctx.get_child()) for step in self._steps)


class ChainOutputs(Operator[T_contra, Iterator[T_co]]):
    """Chains multiple operators together."""

    output_type = Iterator  # type: ignore

    def __init__(self, steps: Operator[T_contra, Iterable[Iterable[T_co]]]) -> None:
        super().__init__()
        self._steps = steps

    def _transform(self, ctx: Context, input_: T_contra) -> Iterator[T_co]:
        return itertools.chain.from_iterable(self._steps(input_, ctx))



""" Generic flexible aggregation step usable in the pipeline
"""
import collections
from typing import Callable, ChainMap, Iterable, Literal, Mapping, TypeVar, Union

from configmate.base import exceptions, operators, registry

T = TypeVar("T")
U = TypeVar("U")
T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)

AggregationSpec = Union[Callable[[Iterable[T_contra]], T_co], Literal["overlay"]]
_SpecT_co = TypeVar("_SpecT_co", bound=AggregationSpec, covariant=True)
AggregatorFactoryMethod = Callable[[_SpecT_co], "Aggregator"]


###
# base class for aggregation steps
###
class Aggregator(operators.Operator[Iterable[T_contra], T_co]):
    """An aggregator that can be used to validate ensure value."""

    input_type = Iterable  # type: ignore


###
# factory for validation strategies
###
class AggregatorFactory(
    registry.StrategyRegistryMixin[AggregationSpec, AggregatorFactoryMethod],
):
    @classmethod
    def build_aggregator(cls, key: AggregationSpec[T, U]) -> Aggregator[T, U]:
        return cls.get_first_match(key)(key)


###
# concrete validation steps
###
class FunctionAggregator(Aggregator[T_contra, T_co]):
    def __init__(
        self, aggregator_function: Callable[[Iterable[T_contra]], T_co]
    ) -> None:
        super().__init__()
        self._method = aggregator_function

    def _transform(self, ctx: operators.Context, input_: Iterable[T_contra]) -> T_co:
        return self._method(input_)


class OverlayAggregator(Aggregator[Mapping, ChainMap]):
    output_type = ChainMap

    def _transform(self, ctx: operators.Context, input_: Iterable[Mapping]) -> ChainMap:
        chainmap = self._make_chainmap(input_)
        chainmap.maps.reverse()  # last map has highest priority
        return chainmap

    def _make_chainmap(self, configs: Iterable) -> collections.ChainMap:
        try:
            configs = list(configs)
            return collections.ChainMap(*map(dict, configs))
        except (ValueError, TypeError) as exc:
            raise exceptions.AggregationFailure(f"Can't aggregate {configs=}") from exc


###
# register strategies in order of priority
###
AggregatorFactory.register(callable, FunctionAggregator)
AggregatorFactory.register(lambda spec: spec == "overlay", OverlayAggregator)



""" Filereading step for the pipeline
"""
import pathlib

from configmate.base import operators, types


class FileReader(operators.Operator[types.FilePath, str]):
    def __init__(self, encoding: str = "utf-8") -> None:
        super().__init__()
        self.encoding = encoding

    def _transform(self, ctx: operators.Context, input_: types.FilePath) -> str:
        return pathlib.Path(input_).read_text(encoding=self.encoding)



""" Generic flexible interpolation steps usable in the pipeline
"""
import contextlib
import re
from typing import Callable, Iterable, Mapping, Set, TypeVar, Union

from configmate.base import constants, operators, registry

InterpolatorSpec = Union[
    Callable[[str], str],
    Mapping[str, str],
    Iterable["InterpolatorSpec"],
]
_SpecT_co = TypeVar("_SpecT_co", bound=InterpolatorSpec, covariant=True)
InterpolatorFactoryMethod = Callable[[_SpecT_co], "StringInterpolator"]


###
# base class for interpolation steps
###
class StringInterpolator(operators.Operator[str, str]):
    input_type = str
    output_type = str


###
# factory for validation strategies
###
class InterpolatorFactory(
    registry.StrategyRegistryMixin[InterpolatorSpec, InterpolatorFactoryMethod],
):
    @classmethod
    def build_interpolator(cls, key: InterpolatorSpec) -> StringInterpolator:
        return cls.get_first_match(key)(key)


###
# concrete interpolation steps
###
class FunctionalInterpolator(StringInterpolator):
    def __init__(self, interpolation_method: Callable[[str], str]) -> None:
        super().__init__()
        self._method = interpolation_method

    def _transform(self, ctx: operators.Context, input_: str) -> str:
        return self._method(input_)


class VariableInterpolator(StringInterpolator):
    def __init__(
        self,
        substitutions: Mapping[str, str] = constants.ENVIRONMENT,
        pattern: re.Pattern = constants.BASH_VAR_PATTERN,
    ) -> None:
        super().__init__()
        self._substitutions = substitutions
        self._sub_pattern = pattern

    def _transform(self, ctx: operators.Context, input_: str) -> str:
        missing_vars: Set[str] = set()

        def replacer(match: re.Match) -> str:
            nonlocal missing_vars
            env_var_name = match.group("variable")
            with contextlib.suppress(KeyError):
                return self._substitutions[env_var_name]
            missing_vars.add(env_var_name)
            return match.group("default_value") or match.group(0)

        ctx[self].missing_vars = missing_vars
        return self._sub_pattern.sub(replacer, input_)


class InterpolatorChain(StringInterpolator):
    build_interpolator = InterpolatorFactory.build_interpolator

    def __init__(self, interpolators: Iterable[operators.Operator[str, str]]) -> None:
        super().__init__()
        if not (interpolators := list(interpolators)):
            raise ValueError("At least one interpolator is required for a chain.")
        self._interpolation = interpolators.pop(0)
        for interpolator in interpolators:
            self._interpolation = self._interpolation.pipe_to(interpolator)

    def _transform(self, ctx: operators.Context, input_: str) -> str:
        return self._interpolation(input_, ctx)

    @classmethod
    def from_factory(cls, spec: Iterable[InterpolatorSpec]) -> "InterpolatorChain":
        spec = list(spec)  # ensures we dont exhaust the iterator
        if all(callable(s) or isinstance(s, Mapping) for s in spec):
            return cls(map(cls.build_interpolator, spec))
        raise ValueError("Only callables and maps supported for chained interpolation.")


###
# register concrete interpolation steps in order of preference
###
# fmt: off
InterpolatorFactory.register(callable, FunctionalInterpolator)
InterpolatorFactory.register(lambda s: isinstance(s, Mapping), VariableInterpolator)
InterpolatorFactory.register(lambda s: isinstance(s, Iterable), InterpolatorChain.from_factory)
# fmt: on



""" Components for config pipelines
"""
from configmate.components.aggregators import (
    AggregationSpec,
    AggregatorFactory,
    FunctionAggregator,
    OverlayAggregator,
)
from configmate.components.cli_readers import (
    ArgSelector,
    CliSectionReader,
    KeyValueSelector,
)
from configmate.components.interpolators import (
    FunctionalInterpolator,
    InterpolatorChain,
    InterpolatorFactory,
    InterpolatorSpec,
    VariableInterpolator,
)
from configmate.components.parsers import (
    FileFormatParserRegistry,
    FunctionParser,
    InferredParser,
    IniParser,
    JsonParser,
    ParserFactory,
    ParsingSpec,
    TomlParser,
    XmlParser,
    YamlParser,
)
from configmate.components.selectors import (
    SectionSelectionSpec,
    SectionSelector,
    SectionSelectorFactory,
)
from configmate.components.validators import (
    FunctionValidator,
    TypeValidatorFactory,
    ValidationSpec,
)



import pathlib
from typing import Iterable, Iterator, Optional, Sequence, Tuple, TypeVar, Union

from configmate.base import operators, types

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)
U = TypeVar("U")
U_contra = TypeVar("U_contra", contravariant=True)
V = TypeVar("V")


def compose_cli_processor(
    section_reader: operators.Operator[types.CliArgs, Sequence[str]],
    filepath_selector: operators.Operator[types.CliArgs, Iterable[types.FilePath]],
    file_parser: operators.Operator[types.FilePath, T_co],
    kwarg_selector: operators.Operator[types.CliArgs, Iterable[Tuple[str, str]]],
    kwarg_parser: operators.Operator[Tuple[str, str], T_co],
) -> operators.Operator[types.CliArgs, Iterator[T_co]]:
    return section_reader.pipe_to(
        operators.ChainOutputs(  # itertools.chain over the outputs of the following:
            operators.JoinOutputs(  # join: parsed files, then parsed args
                filepath_selector.pipe_to(operators.MapIterable(file_parser)),
                kwarg_selector.pipe_to(operators.MapIterable(kwarg_parser)),
            )
        )
    )


def compose_file_processor(
    path_validator: operators.Operator[types.FilePath, pathlib.Path],
    file_reader: operators.Operator[pathlib.Path, str],
    interpolator: Optional[operators.Operator[str, str]],
    parser: operators.Operator[str, T],
    section_selector: Optional[operators.Operator[T, U]],
) -> Union[
    operators.Pipeline[types.FilePath, T], operators.Pipeline[types.FilePath, U]
]:
    return (
        path_validator.pipe_to(file_reader)  # reads in the file as a string
        .pipe_to(interpolator)  # OPTIONAL: interpolate the string file
        .pipe_to(parser)  # parse the file into an object
        .pipe_to(section_selector)  # OPTIONAL: select the section from the config
    )


def compose_config_merger(
    aggregator: operators.Operator[Iterable[T_contra], U_contra],
    validator: Optional[operators.Operator[U_contra, V]],
) -> Union[
    operators.Operator[Iterable[T_contra], U_contra],
    operators.Pipeline[Iterable[T_contra], V],
]:
    return aggregator.pipe_to(validator)  # validator may be none



""" Parsing steps for the pipeline
"""
import configparser
import dataclasses
import json
import os
import pathlib
from typing import Any, Callable, Dict, Generic, Literal, Type, TypeVar, Union
from xml.etree import ElementTree as etree

import toml
import yaml

from configmate.base import operators, registry, types

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
SpecT_co = TypeVar("SpecT_co", bound="ParsingSpec", covariant=True)


###
# Parser base class
###
class Parser(operators.Operator[Any, T_co], Generic[T_co]):
    ...


@dataclasses.dataclass
class InferFrom(Generic[T_co]):
    source_operator: T_co


ParsingSpec = Union[
    Callable[[str], T_co],
    Union[str, os.PathLike],
    InferFrom[operators.Operator[Any, types.FilePath]],
]
ParserFactoryMethod = Callable[[SpecT_co], Parser[T_co]]


###
# factory for parser strategies
###
class ParserFactory(registry.StrategyRegistryMixin[ParsingSpec, ParserFactoryMethod]):
    @classmethod
    def infer_from(cls, step: operators.Operator[Any, types.FilePath]) -> Parser[Any]:
        return cls.build_parser(InferFrom(step))

    @classmethod
    def build_parser(cls, key: ParsingSpec[T]) -> Parser[T]:
        return cls.get_first_match(key)(key)


###
# factory for file format parser strategies
###
class FileFormatParserRegistry(registry.DictRegistryMixin[str, Type[Parser[Any]]]):
    @classmethod
    def infer_parser(cls, path_or_extension: types.FilePath) -> Parser[Any]:
        if suffix := pathlib.Path(path_or_extension).suffix:
            return cls.lookup(suffix)()
        return cls.lookup(cls._normalize_extension(path_or_extension))()

    @classmethod
    def add_strategy(cls, strategy: Type[Parser[Any]], *extensions: str) -> None:
        for extension in extensions:
            cls.register(cls._normalize_extension(extension), strategy)

    @staticmethod
    def _normalize_extension(extension: Union[str, os.PathLike]) -> str:
        return f".{str(extension).lstrip('.')}"


###
# concrete parsers
###
class FunctionParser(Parser[T_co]):
    def __init__(self, parser: Callable[[str], T_co]) -> None:
        super().__init__()
        self._parser = parser

    def _transform(self, ctx: operators.Context, input_: Any) -> T_co:
        return self._parser(input_)


class InferredParser(Parser[Any]):
    """A parser that infers the file format from the file extension."""

    infer_parser = staticmethod(FileFormatParserRegistry.infer_parser)

    def __init__(
        self, infer_via: InferFrom[operators.Operator[Any, types.FilePath]]
    ) -> None:
        super().__init__()
        self._path_sender = infer_via.source_operator
        infer_via.source_operator.append_callback(self._store_path)

    def _transform(self, ctx: operators.Context, input_: str) -> Any:
        filename: types.FilePath = ctx[self._path_sender].filepath
        compatible_parser = self.infer_parser(filename)
        return compatible_parser(input_)

    def _store_path(self, ctx: operators.Context, result: types.FilePath) -> None:
        ctx[self._path_sender].filepath = result


###
# File format specific parsers
###
class JsonParser(Parser[Any]):
    def _transform(self, ctx: operators.Context, input_: Any) -> Any:
        return json.loads(input_)


class IniParser(Parser[Any]):
    def _transform(self, ctx: operators.Context, input_: Any) -> Any:
        (cnfparser := configparser.ConfigParser()).read_string(input_)
        return self._convert_ini_to_dict(cnfparser)

    @staticmethod
    def _convert_ini_to_dict(cnfparser: configparser.ConfigParser) -> Dict[str, Dict]:
        return {section: dict(cnfparser[section]) for section in cnfparser.sections()}


class XmlParser(Parser[Any]):
    XmlTree = Union[Literal[None], str, Dict[str, "XmlTree"]]

    def _transform(self, ctx: operators.Context, input_: Any) -> Any:
        root = etree.fromstring(input_)
        return self._convert_etree_to_dict(root)

    @staticmethod
    def _convert_etree_to_dict(element: etree.Element) -> XmlTree:
        if not (children := list(element)):
            return element.text
        return {
            child.tag: child.text
            if not list(child)
            else XmlParser._convert_etree_to_dict(child)
            for child in children
        }


class YamlParser(Parser[Any]):
    def _transform(self, ctx: operators.Context, input_: Any) -> Any:
        return yaml.safe_load(input_)


class TomlParser(Parser[Any]):
    def _transform(self, ctx: operators.Context, input_: Any) -> Any:
        return toml.loads(input_)


###
# register default strategies
###
# fmt: off
FileFormatParserRegistry.add_strategy(JsonParser, ".json", ".JSON")
FileFormatParserRegistry.add_strategy(IniParser, ".ini", ".INI")
FileFormatParserRegistry.add_strategy(XmlParser, ".xml", ".XML")
FileFormatParserRegistry.add_strategy(TomlParser, ".tml", ".toml", ".TML", ".TOML")
FileFormatParserRegistry.add_strategy(YamlParser, '.yml', ".yaml", '.YML', ".YAML")

ParserFactory.register(lambda spec: isinstance(spec, InferFrom), InferredParser)
ParserFactory.register(callable, FunctionParser)
ParserFactory.register(lambda spec: isinstance(spec, str), FileFormatParserRegistry.infer_parser)
# fmt: on



import collections
from typing import Any, Callable, Deque, Dict, Generic, NoReturn, Tuple, TypeVar

from configmate.base import exceptions, types

T = TypeVar("T")
T_contra = TypeVar("T_contra", contravariant=True)


class DictRegistryMixin(types.HasDescription, Generic[T_contra, T]):
    _registry: Dict[T_contra, T]

    def __init_subclass__(cls, *args, **kwargs) -> None:
        cls._registry = {}
        return super().__init_subclass__(*args, **kwargs)

    @classmethod
    def lookup(cls, key: T_contra) -> T:
        """Retrieves first triggered strategy or raises `~exceptions.NoApplicableStrategy`."""
        if key in cls._registry:
            return cls._registry[key]
        _raise_missing(cls, key)

    @classmethod
    def register(cls, key: T_contra, value: T) -> None:
        """Inserts a new strategy at the given rank (position in queue)."""
        cls._registry[key] = value

    @classmethod
    def describe(cls) -> str:
        """Returns a string representation of all items in the registry."""
        return "\n".join(f"{key}: {value}" for key, value in cls._registry.items())


class StrategyRegistryMixin(types.HasDescription, Generic[T_contra, T]):
    _registry: Deque[Tuple[Callable[[T_contra], bool], T]]

    def __init_subclass__(cls, *args, **kwargs) -> None:
        cls._registry = collections.deque()
        return super().__init_subclass__(*args, **kwargs)

    @classmethod
    def get_first_match(cls, key: T_contra) -> T:
        """Retrieves first triggered strategy or raises `~exceptions.NoApplicableStrategy`."""
        for entry in (entry for trigger, entry in cls._registry if trigger(key)):
            return entry
        _raise_missing(cls, key)

    @classmethod
    def register(
        cls, trigger: Callable[[T_contra], bool], entry: T, rank: int = -1
    ) -> None:
        """Inserts a new strategy at the given rank (position in queue)."""
        cls._registry.insert(rank, (trigger, entry))

    @classmethod
    def describe(cls) -> str:
        """Returns a string representation of all items in the registry."""
        return "\n".join(
            f"{entry[1]} (triggered by: {entry[0].__name__})" for entry in cls._registry
        )


def _raise_missing(cls: types.HasDescription, unfound_key: Any) -> NoReturn:
    raise exceptions.NoApplicableStrategy(
        f"Missing strategy for {unfound_key}, please register one."
        f"\n\tAvailable strategies:\n{cls.describe()}"
    )



import functools
import itertools
from typing import Any, Callable, Optional, TypeVar, Union, overload

from configmate.base import constants, operators, types
from configmate.components import (
    aggregators,
    interpolators,
    parsers,
    selectors,
    validators,
)
from configmate.core import builders

T = TypeVar("T")
U = TypeVar("U")


# fmt: off
@overload
def get_config( # if no validation is specified, we get the output of the aggregation
    *config_files: types.FilePath,
    ## file reading
    interpolation: Optional[interpolators.InterpolatorSpec] = constants.ENVIRONMENT,
    parsing: Union[types.Infer, parsers.ParsingSpec] = constants.INFER_FROM_PATH,
    section: Optional[selectors.SectionSelectionSpec] = None,
    aggregation: aggregators.AggregationSpec[T, U] = ...,
    validation: None = None,
    ## CLI overlay options
    cli_section_name: Optional[str] = None,
    cli_section_end: Optional[str] = constants.CLI_SECTION_END_TOKEN,
    cli_overlay_file_prefix: str = constants.CLI_OVERLAY_FILE_PREFIX,
    cli_overlay_arg_key_prefix: str = constants.CLI_OVERLAY_KWARG_KEY_PREFIX,
    cli_overlay_arg_key_delimiter: str = constants.CLI_OVERLAY_KWARG_KEY_DELIMITER,
    cli_overlay_arg_parser: Callable[[str], Any] = constants.CLI_OVERLAY_KWARG_VAL_PARSER,
    ## encoding
    file_encoding: str = constants.SYS_DEFAULT_FILE_ENCODING,
) -> U: ...
@overload
def get_config( # the validation determines the return type
    *config_files: types.FilePath,
    ## file reading
    interpolation: Optional[interpolators.InterpolatorSpec] = constants.ENVIRONMENT,
    parsing: Union[types.Infer, parsers.ParsingSpec] = constants.INFER_FROM_PATH,
    section: Optional[selectors.SectionSelectionSpec] = None,
    aggregation: aggregators.AggregationSpec = constants.OVERLAY,
    validation: validators.ValidationSpec[T, U] = ...,
    ## CLI overlay options
    cli_section_name: Optional[str] = None,
    cli_section_end: Optional[str] = constants.CLI_SECTION_END_TOKEN,
    cli_overlay_file_prefix: str = constants.CLI_OVERLAY_FILE_PREFIX,
    cli_overlay_arg_key_prefix: str = constants.CLI_OVERLAY_KWARG_KEY_PREFIX,
    cli_overlay_arg_key_delimiter: str = constants.CLI_OVERLAY_KWARG_KEY_DELIMITER,
    cli_overlay_arg_parser: Callable[[str], Any] = constants.CLI_OVERLAY_KWARG_VAL_PARSER,
    ## encoding
    file_encoding: str = constants.SYS_DEFAULT_FILE_ENCODING,
) -> U: ...
# fmt: on
def get_config(  # pylint: disable=too-many-arguments,too-many-locals
    *config_files: types.FilePath,
    ## file reading
    interpolation: Optional[interpolators.InterpolatorSpec] = constants.ENVIRONMENT,
    parsing: Union[types.Infer, parsers.ParsingSpec] = constants.INFER_FROM_PATH,
    section: Optional[selectors.SectionSelectionSpec] = None,
    aggregation: aggregators.AggregationSpec = constants.OVERLAY,
    validation: Optional[validators.ValidationSpec] = None,
    ## CLI overlay options
    cli_section_name: Optional[str] = None,
    cli_section_end: Optional[str] = constants.CLI_SECTION_END_TOKEN,
    cli_overlay_file_prefix: str = constants.CLI_OVERLAY_FILE_PREFIX,
    cli_overlay_arg_key_prefix: str = constants.CLI_OVERLAY_KWARG_KEY_PREFIX,
    cli_overlay_arg_key_delimiter: str = constants.CLI_OVERLAY_KWARG_KEY_DELIMITER,
    cli_overlay_arg_parser: Callable[
        [str], Any
    ] = constants.CLI_OVERLAY_KWARG_VAL_PARSER,
    ## encoding
    file_encoding: str = constants.SYS_DEFAULT_FILE_ENCODING,
):
    """Get the config from the given files, with the given options."""
    file_processing_pipeline = builders.build_fileprocessor(
        interpolation=interpolation,
        parsing=parsing,
        section=section,
        file_encoding=file_encoding,
    )
    cli_reader = builders.build_cli_reader(
        section_name=cli_section_name,
        section_end=cli_section_end,
        file_arg_prefix=cli_overlay_file_prefix,
        cli_overlay_file_parser=file_processing_pipeline,
        kwarg_key_delimiter=cli_overlay_arg_key_delimiter,
        kwarg_key_prefix=cli_overlay_arg_key_prefix,
        kwarg_value_parser=cli_overlay_arg_parser,
    )
    config_merger = builders.build_config_merger(
        aggregation=aggregation,
        validation=validation,
    )
    return config_merger(
        itertools.chain(
            operators.MapIterable(file_processing_pipeline)(config_files),
            cli_reader(constants.CLI_ARGS),  # pylint: disable=not-callable
        )
    )


def configure(  # pylint: disable=too-many-arguments,too-many-locals
    *config_files: types.FilePath,
    ## file reading
    interpolation: Optional[interpolators.InterpolatorSpec] = constants.ENVIRONMENT,
    parsing: Union[types.Infer, parsers.ParsingSpec] = constants.INFER_FROM_PATH,
    section: Optional[selectors.SectionSelectionSpec] = None,
    aggregation: aggregators.AggregationSpec = constants.OVERLAY,
    validation: Optional[validators.ValidationSpec] = None,
    ## CLI overlay options
    cli_section_name: Optional[str] = None,
    cli_section_end: Optional[str] = constants.CLI_SECTION_END_TOKEN,
    cli_overlay_file_prefix: str = constants.CLI_OVERLAY_FILE_PREFIX,
    cli_overlay_arg_key_prefix: str = constants.CLI_OVERLAY_KWARG_KEY_PREFIX,
    cli_overlay_arg_key_delimiter: str = constants.CLI_OVERLAY_KWARG_KEY_DELIMITER,
    cli_overlay_arg_parser: Callable[
        [str], Any
    ] = constants.CLI_OVERLAY_KWARG_VAL_PARSER,
    ## encoding
    file_encoding: str = constants.SYS_DEFAULT_FILE_ENCODING,
) -> Callable[[Callable[..., U]], Callable[..., U]]:
    """Decorator to configure a function with the given config."""
    file_processing_pipeline = builders.build_fileprocessor(
        interpolation=interpolation,
        parsing=parsing,
        section=section,
        file_encoding=file_encoding,
    )
    cli_reader = builders.build_cli_reader(
        section_name=cli_section_name,
        section_end=cli_section_end,
        file_arg_prefix=cli_overlay_file_prefix,
        cli_overlay_file_parser=file_processing_pipeline,
        kwarg_key_delimiter=cli_overlay_arg_key_delimiter,
        kwarg_key_prefix=cli_overlay_arg_key_prefix,
        kwarg_value_parser=cli_overlay_arg_parser,
    )
    config_merger = builders.build_config_merger(
        aggregation=aggregation,
        validation=validation,
    )

    def decorator(func: Callable[..., U]) -> Callable[..., U]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> U:
            kwargs.update(dict(zip(func.__code__.co_varnames, args)))
            new_kwargs = config_merger(
                itertools.chain(
                    operators.MapIterable(file_processing_pipeline)(config_files),
                    cli_reader(constants.CLI_ARGS),  # pylint: disable=not-callable
                    (kwargs,),
                )
            )
            return func(**new_kwargs)

        return wrapper

    return decorator



""" Generic flexible aggregation step usable in the pipeline
"""
import functools

from typing import Any, Callable, NoReturn, Sequence, TypeVar, Union

from configmate.base import exceptions, operators, registry

T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)
SpecT_co = TypeVar("SpecT_co", bound="SectionSelectionSpec", covariant=True)


class SectionSelector(operators.Operator[T_contra, T_co]):
    """Selects a section from the config"""


SectionSelectionSpec = Union[Callable[[T_contra], T_co], str, Sequence[str]]
SectionSelectorFactoryMethod = Callable[[SpecT_co], SectionSelector[T_contra, T_co]]


###
# factory for validation strategies
###
class SectionSelectorFactory(
    registry.StrategyRegistryMixin[SectionSelectionSpec, SectionSelectorFactoryMethod]
):
    @classmethod
    def build_selector(
        cls, key: SectionSelectionSpec[T_contra, T_co]
    ) -> SectionSelector[T_contra, T_co]:
        return cls.get_first_match(key)(key)


###
# concrete validation steps
###
class FunctionSectionSelector(SectionSelector):
    def __init__(self, selector_function: Callable[[Any], Any]) -> None:
        super().__init__()
        self._method = selector_function

    def _transform(self, ctx: operators.Context, input_: Any) -> Any:
        return self._method(input_)


class KeySelector(SectionSelector):
    def __init__(self, section: Union[str, Sequence[str]]) -> None:
        super().__init__()
        self._section = (section,) if isinstance(section, str) else section

    def _transform(self, ctx: operators.Context, input_: Any) -> Any:
        try:
            return functools.reduce(lambda d, key: d[key], self._section, input_)
        except (TypeError, KeyError) as exc:
            self._raise_missing(input_, exc)

    def _raise_missing(self, input_: Any, exc: Exception) -> NoReturn:
        raise exceptions.SectionNotFound(f"{self._section=} not in {input_=}") from exc


###
# register strategies in order of priority
###
# fmt: off
SectionSelectorFactory.register(callable, FunctionSectionSelector)
SectionSelectorFactory.register(lambda spec: isinstance(spec, (Sequence, str)), KeySelector)
# fmt: on



import json
import os
import re
import sys
import types
from typing import Literal
from configmate.base import types as configmate_types

SYS_DEFAULT_FILE_ENCODING = sys.getdefaultencoding()

CLI_OVERLAY_FILE_PREFIX = "+"
CLI_OVERLAY_KWARG_KEY_PREFIX = "++"
CLI_OVERLAY_KWARG_KEY_DELIMITER = "."
CLI_OVERLAY_KWARG_VAL_PARSER = json.loads
CLI_SECTION_END_TOKEN = "/"

CLI_ARGS = sys.argv
ENVIRONMENT = types.MappingProxyType(os.environ)  # read-only view of the env-vars
BASH_VAR_PATTERN = re.compile(r"\${(?P<variable>\w+)(?::(?P<default_value>[^}:]+))?}")

OVERLAY: Literal["overlay"] = "overlay"

INFER_FROM_PATH = configmate_types.Infer()



import pathlib
from typing import Callable, Iterable, Iterator, Optional, TypeVar, Union

from configmate.base import constants, operators, types
from configmate.components import (
    aggregators,
    cli_readers,
    filereader,
    interpolators,
    parsers,
    selectors,
    validators,
)
from configmate.core import composers

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)
U_contra = TypeVar("U_contra", contravariant=True)
U = TypeVar("U")
V = TypeVar("V")

build_interpolator = interpolators.InterpolatorFactory.build_interpolator
build_parser = parsers.ParserFactory.build_parser
build_runtime_inferred_parser = parsers.ParserFactory.infer_from
build_config_section_selector = selectors.SectionSelectorFactory.build_selector
build_aggregator = aggregators.AggregatorFactory.build_aggregator
build_validator = validators.TypeValidatorFactory.build_validator


def build_cli_reader(
    cli_overlay_file_parser: operators.Operator[types.FilePath, T_co],
    file_arg_prefix: str = constants.CLI_OVERLAY_FILE_PREFIX,
    section_name: Optional[str] = None,
    section_end: Optional[str] = constants.CLI_SECTION_END_TOKEN,
    kwarg_key_prefix: str = constants.CLI_OVERLAY_KWARG_KEY_PREFIX,
    kwarg_key_delimiter: str = constants.CLI_OVERLAY_KWARG_KEY_DELIMITER,
    kwarg_value_parser: Callable[[str], T] = constants.CLI_OVERLAY_KWARG_VAL_PARSER,
) -> operators.Operator[types.CliArgs, Iterator[Union[types.NestedDict[T], T_co]]]:
    return composers.compose_cli_processor(
        section_reader=cli_readers.CliSectionReader(section_name, section_end),
        filepath_selector=cli_readers.ArgSelector(file_arg_prefix),
        file_parser=cli_overlay_file_parser,
        kwarg_selector=cli_readers.KeyValueSelector(kwarg_key_prefix),
        kwarg_parser=cli_readers.DictParser(kwarg_key_delimiter, kwarg_value_parser),
    )


def build_fileprocessor(
    interpolation: Optional[interpolators.InterpolatorSpec] = constants.ENVIRONMENT,
    parsing: Union[types.Infer, parsers.ParsingSpec[T]] = constants.INFER_FROM_PATH,
    section: Optional[selectors.SectionSelectionSpec[T, U]] = None,
    file_encoding: str = constants.SYS_DEFAULT_FILE_ENCODING,
) -> Union[
    operators.Pipeline[types.FilePath, T], operators.Pipeline[types.FilePath, U]
]:
    return composers.compose_file_processor(
        path_validator=(path_factory := validators.FunctionValidator(pathlib.Path)),
        file_reader=filereader.FileReader(encoding=file_encoding),
        interpolator=(
            build_interpolator(interpolation) if interpolation is not None else None
        ),
        parser=(
            build_runtime_inferred_parser(path_factory)
            if isinstance(parsing, types.Infer)
            else build_parser(parsing)
        ),
        section_selector=(
            build_config_section_selector(section) if section is not None else None
        ),
    )


def build_config_merger(
    aggregation: aggregators.AggregationSpec[T_contra, U],
    validation: Optional[validators.ValidationSpec[U, V]] = None,
) -> Union[
    operators.Operator[Iterable[T_contra], U],
    operators.Pipeline[Iterable[T_contra], V],
]:
    return composers.compose_config_merger(
        build_aggregator(aggregation),
        build_validator(validation) if validation is not None else None,
    )



""" Generic flexible validation step usable in the pipeline
"""

from typing import Any, Callable, Type, TypeVar, Union

from configmate.base import operators, registry

T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)
SpecT_contra = TypeVar("SpecT_contra", bound="ValidationSpec", contravariant=True)


class TypeValidator(operators.Operator[T_contra, T_co]):
    """A validator that can be used to validate ensure value."""


ValidationSpec = Union[Callable[[T_contra], T_co], Type[T_co]]
ValidatorFactoryMethod = Callable[[SpecT_contra], TypeValidator[T_contra, T_co]]


###
# factory for validation strategies
###
class TypeValidatorFactory(
    registry.StrategyRegistryMixin[ValidationSpec, ValidatorFactoryMethod]
):
    @classmethod
    def build_validator(
        cls, key: ValidationSpec[T_contra, Union[T_co, Any]]
    ) -> TypeValidator[T_contra, T_co]:
        return cls.get_first_match(key)(key)


###
# concrete validation steps
###
class FunctionValidator(TypeValidator[T_contra, T_co]):
    def __init__(self, validator_method: Callable[[T_contra], T_co]) -> None:
        super().__init__()
        self._method = validator_method

    def _transform(self, ctx: operators.Context, input_: Any) -> T_co:
        return self._method(input_)


###
# register strategies in order of priority
###
TypeValidatorFactory.register(callable, FunctionValidator)



"""This module contains all the exceptions that are raised by configmate."""

# fmt: off
class ConfigmateError(Exception): "Base class for all configmate errors."

class NoApplicableStrategy(ConfigmateError): "Raised when no strategy is applicable."
class SectionNotFound(ConfigmateError): "Raised when the specified section is not found."
class MissingEnvironmentVariable(ConfigmateError): "Raised for missing environment variables."
class AggregationFailure(ConfigmateError): "Raised when aggregation fails."
# fmt: on



import os
import pathlib
from typing import Mapping, Protocol, Sequence, TypeVar, Union

T = TypeVar("T")

FilePathTypes = (str, os.PathLike, pathlib.Path)
FilePath = Union[str, os.PathLike, pathlib.Path]
CliArgs = Sequence[str]
NestedDict = Mapping[str, Union[T, "NestedDict[T]"]]

# fmt: off
class Unset: 'Used to mark a value that will be set.'
class Infer: 'Marks a value that is inferred during runtime.'
class Sentinel: 'Used to catch exhausted iterators.'
SENTINEL = Sentinel()
# fmt: on


class HasDescription(Protocol):
    @classmethod
    def describe(cls) -> str:
        ...
